/***************************************************************************
 *   Copyright (C) 2005-2006 by Georg Hennig                               *
 *   Email: georg.hennig@web.de                                            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <qfile.h>

#include <kapplication.h>
#include <klocale.h>
#include <kmdcodec.h>
#include <kmessagebox.h>
#include <ktempfile.h>

#include "komparejob.h"
#include "komparatorwidget.h"

#include "kfileitemext.h"

#include "config.h"

KFileItemExt::KFileItemExt( const KIO::UDSEntry &_entry, const KURL &_url,
                            bool _determineMimeTypeOnDemand, bool _urlIsDirectory ) :
  KFileItem( _entry, _url, _determineMimeTypeOnDemand, _urlIsDirectory )
{
  md5 = NULL;
  hasdupes = 0;
  isdupe = 0;
  isdupe_comp = 0;
  duplicates = NULL;
  dup_parent = NULL;
  brother = NULL;
  next = NULL;
  tmpFile = NULL;
}

KFileItemExt::KFileItemExt( mode_t _mode, mode_t _permissions,
                            const KURL &_url, bool _determineMimeTypeOnDemand ) :
  KFileItem( _mode, _permissions, _url, _determineMimeTypeOnDemand )
{
  md5 = NULL;
  hasdupes = 0;
  isdupe = 0;
  isdupe_comp = 0;
  duplicates = NULL;
  dup_parent = NULL;
  brother = NULL;
  next = NULL;
  tmpFile = NULL;
}

KFileItemExt::KFileItemExt( const KURL &url, const QString &mimeType, mode_t mode ) :
  KFileItem( url, mimeType, mode )
{
  md5 = NULL;
  hasdupes = 0;
  isdupe = 0;
  isdupe_comp = 0;
  duplicates = NULL;
  dup_parent = NULL;
  brother = NULL;
  next = NULL;
  tmpFile = NULL;
}

KFileItemExt::KFileItemExt( const KFileItem &item ) :
  KFileItem( item )
{
  md5 = NULL;
  hasdupes = 0;
  isdupe = 0;
  isdupe_comp = 0;
  duplicates = NULL;
  dup_parent = NULL;
  brother = NULL;
  next = NULL;
  tmpFile = NULL;
}

KFileItemExt::KFileItemExt( const KFileItemExt &item ) :
  KFileItem( item )
{
  dir = item.dir;
  hasdupes = item.hasdupes;
  isdupe = item.isdupe;
  duplicates = item.duplicates;
  dup_parent = item.dup_parent;
  brother = item.brother;
  next = item.next;
  md5 = item.md5;
  tmpFile = NULL;
}

KFileItemExt::~KFileItemExt()
{
  if ( md5 != NULL )
  {
    delete md5;
    md5=NULL;
  }

  if ( next != NULL ) // We don't need to delete the duplicates, as duplicates also appera in "next".
  {
    delete next;
    next=NULL;
  }
}

QCString KFileItemExt::MD5( KomparatorWidget *parent )
{
  if ( !md5 )
  {
    const int BUFFER_SIZE = 512*1024;
    char buffer[BUFFER_SIZE];
    int read_data;

    if ( isLocalFile() )
    {
      QFile file;
      file.setName( url().directory( FALSE )+url().fileName() );
      if ( !file.open( IO_Raw | IO_ReadOnly ) )
      {
        KMessageBox::error( 0,
          i18n( "Error getting md5 sum of local file %1." ).arg( url().directory( FALSE )+url().fileName() ),
          i18n( "Error while comparing files" ) );
        return QCString( "d41d8cd98f00b204e9800998ecf8427e" ); // checksum of empty file.
      }

      md5 = new KMD5();
      while ( ( read_data = file.readBlock( buffer, BUFFER_SIZE ) ) > 0 )
      {
        md5->update( buffer, read_data );

        if ( parent->kompare->cancel() ) break; // md5 sum will be invalid, but cancel will delete it anyway.
      }

      file.close();
    }
    else
    {
      tmpFile = NULL;
      download_error = -1;

      mutex.lock();

      QCustomEvent *event = new QCustomEvent( DOWNLOAD_FILE );
      event->setData( this );
      QApplication::postEvent( parent, event );

      while ( download_error == -1 )
      {
        mutex.unlock();
        usleep( 50 );
        mutex.lock();
      }

      mutex.unlock();

      if ( parent->kompare->cancel() )
      {
        tmpFile->close();
        tmpFile->unlink();

        QCustomEvent *event2 = new QCustomEvent( DELETE_FILE );
        event2->setData( tmpFile );
        QApplication::postEvent( parent, event2 );

        tmpFile = NULL;

        return QCString( "d41d8cd98f00b204e9800998ecf8427e" ); // checksum of empty file.
      }

      md5 = new KMD5();

      if ( download_error == 0 ) // everything went well.
      {
        QFile file( tmpFile->name() );
        tmpFile->close();
        if ( !file.open( IO_Raw | IO_ReadOnly ) )
        {
          KMessageBox::sorry( NULL, i18n( "Opening temporary file %1 failed. MD5 checksum won't be correct." ).arg( tmpFile->name() ) );
        }
        else
        {
          while ( ( read_data = file.readBlock( buffer, BUFFER_SIZE ) ) > 0 )
          {
            md5->update( buffer, read_data );

            if ( parent->kompare->cancel() ) break; // md5 sum will be invalid, but cancel will delete it anyway.
          }

          file.close();
        }
      }
      else // download errors are handled by komparatorwidget.
      {
        tmpFile->close();
      }
      tmpFile->unlink();

      QCustomEvent *event2 = new QCustomEvent( DELETE_FILE );
      event2->setData( tmpFile );
      QApplication::postEvent( parent, event2 );

      tmpFile = NULL;
    }
  }

  return md5->hexDigest();
}

QString KFileItemExt::getFile( KomparatorWidget *parent )
{
  QString ret = "";

  if ( isLocalFile() )
  {
    ret = url().directory( FALSE ) + url().fileName();
  }
  else
  {
    tmpFile = NULL;
    download_error = -1;

    mutex.lock();

    QCustomEvent *event = new QCustomEvent( DOWNLOAD_FILE );
    event->setData( this );
    QApplication::postEvent( parent, event );

    while ( download_error == -1 ) // wait until the KIO::Job has finished (in komparatorwidget).
    {
      mutex.unlock();
      usleep( 50 );
      mutex.lock();
    }

    mutex.unlock();

    if ( parent->kompare->cancel() )
    {
      tmpFile->close();
      tmpFile->unlink();

      QCustomEvent *event2 = new QCustomEvent( DELETE_FILE );
      event2->setData( tmpFile );
      QApplication::postEvent( parent, event2 );

      tmpFile = NULL;

      return "";
    }

    if ( download_error == 0 ) // everything went well. otherwise will fail in komparejob.
    {
      ret = tmpFile->name();
    }

    tmpFile->close();
  }

  return ret;
}

void KFileItemExt::deleteFile( KomparatorWidget *parent )
{
  if ( tmpFile != NULL )
  {
    tmpFile->unlink();

    QCustomEvent *event = new QCustomEvent( DELETE_FILE );
    event->setData( tmpFile );
    QApplication::postEvent( parent, event );

    tmpFile = NULL;
  }
}

void KFileItemExt::setTmpFile( KTempFile *_tmpFile )
{
  mutex.lock();
  tmpFile = _tmpFile;
  mutex.unlock();
}

void KFileItemExt::setDownloadError( int _download_error )
{
  mutex.lock();
  download_error = _download_error;
  mutex.unlock();
}
