/***************************************************************************
 *   Copyright (C) 2005-2006 by Georg Hennig                               *
 *   Email: georg.hennig@web.de                                            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


/* This file uses main parts of "fdupes.c" from fdupes-1.40.
   Copyright: */

/* FDUPES Copyright (c) 1999 Adrian Lopez

   Permission is hereby granted, free of charge, to any person
   obtaining a copy of this software and associated documentation files
   (the "Software"), to deal in the Software without restriction,
   including without limitation the rights to use, copy, modify, merge,
   publish, distribute, sublicense, and/or sell copies of the Software,
   and to permit persons to whom the Software is furnished to do so,
   subject to the following conditions:

   The above copyright notice and this permission notice shall be
   included in all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
   OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY 
   CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, 
   TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
   SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. */

#include <qapplication.h>
#include <qfile.h>

#include <klocale.h>
#include <kmdcodec.h>
#include <kmessagebox.h>

#include "komparator.h"
#include "komparatorwidget.h"

#include "komparejob.h"

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

KompareJob::KompareJob( QWidget *_parent )
  : parent( _parent )
{
  files_dupes = NULL;
  files_comp = NULL;
  ignore_empty = FALSE;
  binary_comparison = FALSE;
  initialized = FALSE;
}

bool KompareJob::cancel()
{
  mutex.lock();
  bool ret = canceled;
  mutex.unlock();
  return ret;
}

void KompareJob::emitStatus( QString str )
{
  QString *new_str = new QString( str );

  QCustomEvent *event = new QCustomEvent( PROGRESS_STATUS );
  event->setData( new_str );
  QApplication::postEvent( parent, event );
}

void KompareJob::emitProgress( QString str, int percent )
{
  QString *new_str = new QString( str );
  int *new_percent = new int;
  *new_percent = percent;

  QPair< QString *, int * > *resultpair = new QPair< QString *, int * >;

  *resultpair = qMakePair( new_str, new_percent );

  QCustomEvent *event = new QCustomEvent( PROGRESS_PROGRESS );
  event->setData( resultpair );
  QApplication::postEvent( parent, event );
}

void KompareJob::initialize( KFileItemExt *_files_dupes, KFileItemExt *_files_comp, bool _ignore_empty,
                             bool _size, bool _calculate_checksum, bool _binary_comparison,
                             uint _binary_comparision_threshold, uint _number_of_files, KURL _url1, KURL _url2,
                             bool _enable_duplicates_search, bool _enable_missing_search, bool _enable_newer_search )
{
  mutex.lock();

  files_dupes = _files_dupes;
  files_comp = _files_comp;
  ignore_empty = _ignore_empty;
  size = _size;
  calculate_checksum = _calculate_checksum;
  binary_comparison = _binary_comparison;
  binary_comparision_threshold = _binary_comparision_threshold;
  number_of_files = _number_of_files;
  url1 = _url1;
  url2 = _url2;

  enable_duplicates_search = _enable_duplicates_search;
  enable_missing_search = _enable_missing_search;
  enable_newer_search = _enable_newer_search;

  initialized = TRUE;
  canceled = FALSE;

  mutex.unlock();
}

void KompareJob::start()
{
  mutex.lock();
  canceled = FALSE;
  mutex.unlock();

  QThread::start();
}

void KompareJob::stop()
{
  mutex.lock();
  canceled = TRUE;
  mutex.unlock();
}

void KompareJob::run()
{
  if ( !initialized ) return;

  KFileItemExt *curfile_dupes;
  KFileItemExt *curfile_comp;
  KFileItemExt *match_dupes = NULL;
  KFileItemExt *match_comp = NULL;
  fileTree *checktree_dupes = NULL;
  fileTree *checktree_comp = NULL;

  curfile_dupes = files_dupes;
  curfile_comp = files_comp;

  QString status;
  uint current_file;

  if ( size && ( enable_duplicates_search || enable_missing_search || enable_newer_search ) )
  {                                                       // comparing duplicates by size is the minimum we can do.
    status = i18n( "Searching for duplicates..." );

    current_file = 0;

    while ( curfile_dupes )
    {
      if ( cancel() )  break;

      emitProgress( status, (int)((current_file*100)/number_of_files) );

      if ( !checktree_dupes ) 
        registerFile( &checktree_dupes, curfile_dupes );
      else 
        match_dupes = checkMatchDupes( &checktree_dupes, checktree_dupes, curfile_dupes );

      if ( match_dupes != NULL )
      {
        if ( !curfile_dupes->isReadable() )
        {
          curfile_dupes = curfile_dupes->next;
          continue;
        }

        if ( !match_dupes->isReadable() )
        {
          curfile_dupes = curfile_dupes->next;
          continue;
        }

        if ( ignore_empty )  if ( match_dupes->size() == 0 )
        {
          curfile_dupes = curfile_dupes->next;
          continue;
        }

        if ( binary_comparison )
        {
          if ( confirmMatchDupes( match_dupes, curfile_dupes ) )
          {
            match_dupes->hasdupes = 1;
            match_dupes->isdupe = 1;
            curfile_dupes->isdupe = 1;
            if ( match_dupes->duplicates )
            {
              match_dupes->duplicates->isdupe = 1;
              match_dupes->duplicates->dup_parent = curfile_dupes;
            }
            curfile_dupes->duplicates = match_dupes->duplicates;
            match_dupes->duplicates = curfile_dupes;
            curfile_dupes->dup_parent = match_dupes;
          }
        }
        else
        {
          match_dupes->hasdupes = 1;
          match_dupes->isdupe = 1;
          curfile_dupes->isdupe = 1;
          if ( match_dupes->duplicates )
          {
            match_dupes->duplicates->isdupe = 1;
            match_dupes->duplicates->dup_parent = curfile_dupes;
          }
          curfile_dupes->duplicates = match_dupes->duplicates;
          match_dupes->duplicates = curfile_dupes;
          curfile_dupes->dup_parent = match_dupes;
        }
      }

      current_file++;
        curfile_dupes = curfile_dupes->next;
    }

    if ( enable_duplicates_search )
    {
      status = i18n( "Appending duplicates to list view..." );
      current_file = 0;

      while ( files_dupes != NULL )
      {
        if ( cancel() )  break;

        emitProgress( status, (int)((current_file*100)/number_of_files) );

        if ( files_dupes->hasdupes )
        {
          QCustomEvent *event = new QCustomEvent( RESULT_DUPLICATE );
          event->setData( files_dupes );
          QApplication::postEvent( parent, event );
        }

        current_file++;
        files_dupes = files_dupes->next;
      }
    }
  }

  if ( ( enable_duplicates_search && !size ) || enable_missing_search || enable_newer_search )
  {
    status = i18n( "Searching for missing / newer files..." );
    current_file = 0;

    while ( curfile_comp )
    {
      if ( cancel() )  break;

      emitProgress( status, (int)((current_file*100)/number_of_files) );

      if ( !checktree_comp ) 
        registerFile( &checktree_comp, curfile_comp );
      else 
        match_comp = checkMatchComp( &checktree_comp, checktree_comp, curfile_comp );

      if ( match_comp != NULL )
      {
        match_comp->hasdupes = 1;
        match_comp->isdupe_comp = 1;
        curfile_comp->isdupe_comp = 1;
        curfile_comp->duplicates = match_comp->duplicates;
        match_comp->duplicates = curfile_comp;
        curfile_comp->dup_parent = match_comp;
      }

      current_file++;
      curfile_comp = curfile_comp->next;
    }

    KFileItemExt *tmpfile;

    status = i18n( "Appending missing / newer files to list views..." );
    current_file = 0;

    while ( files_comp != NULL )
    {
      if ( cancel() ) break;

      emitProgress( status, (int)((current_file*100)/number_of_files) );

      QPtrList< QPair<KFileItemExt*, KFileItemExt*> > list;

      if ( files_comp->hasdupes )
      {
        QPair<KFileItemExt*, KFileItemExt*> *pair = new QPair<KFileItemExt*, KFileItemExt*>;
        *pair = qMakePair( files_comp, (KFileItemExt*)NULL );
        list.append( pair );

        QPair<KFileItemExt*, KFileItemExt*> *tmppair;
        bool inserted;

        tmpfile = files_comp->duplicates;
        while ( tmpfile != NULL )
        {
          if ( cancel() ) break;

          inserted = FALSE;

          for ( tmppair = list.first(); tmppair; tmppair = list.next() ) // if two paths are identical, we have a pair.
          {                                                              // otherwise it's <item, NULL>.
            if ( tmppair->second == NULL )
            {
              if ( confirmMatchComp( tmpfile, tmppair->first ) )
              {
                tmppair->second = tmpfile;
                inserted = TRUE;
              }
            }
          }

          if ( !inserted )
          {
            QPair<KFileItemExt*, KFileItemExt*> *pair2 = new QPair<KFileItemExt*, KFileItemExt*>;
            *pair2 = qMakePair( files_comp, (KFileItemExt*)NULL );
            list.append( pair2 );
          }

          tmpfile = tmpfile->duplicates;
        }

        for ( tmppair = list.first(); tmppair; tmppair = list.next() )
        {
          if ( tmppair->second == NULL ) // a file that has no equivalent on the other side.
          {
            if ( enable_missing_search )
            {
              QCustomEvent *event = new QCustomEvent( RESULT_COMP_MISSING );
              event->setData( tmppair->first );
              QApplication::postEvent( parent, event );
            }
          }
          else                          // found two identical paths. must check for newer file.
          {
            KFileItemExt *neweritem = newerItem( tmppair->first, tmppair->second );
            if ( neweritem != NULL )    // we have files with the same path+name, but different mod time...
            {
              QPair<KFileItemExt*, KFileItemExt*> *resultpair = new QPair<KFileItemExt*, KFileItemExt*>;

              if ( neweritem == tmppair->first ) *resultpair = qMakePair( tmppair->first, tmppair->second );
              else *resultpair = qMakePair( tmppair->second, tmppair->first );

              // check if files are duplicates.
              bool are_duplicates = FALSE;

              if ( resultpair->first->size() == 0 && resultpair->second->size() == 0 ) are_duplicates = TRUE;

              tmpfile = resultpair->first->brother;
              while( tmpfile && !are_duplicates )
              {
                if ( tmpfile == resultpair->second->brother ) are_duplicates = TRUE;
                tmpfile = tmpfile->duplicates;
              }

              tmpfile = resultpair->second->brother;
              while( tmpfile && !are_duplicates )
              {
                if ( tmpfile == resultpair->first->brother ) are_duplicates = TRUE;
                tmpfile = tmpfile->duplicates;
              }

              if ( enable_newer_search )
              { // if !size we always find newer, not newer_equal
                QCustomEvent *event = new QCustomEvent( are_duplicates ? RESULT_COMP_NEWER_EQUAL : RESULT_COMP_NEWER );
                event->setData( resultpair );
                QApplication::postEvent( parent, event );
              }
              else
              {
                delete resultpair;
              }
            }
            else // ... find files that are different, but have the same timestamp
            {
              if ( cancel() ) break;

              // check if files are duplicates.
              bool are_duplicates = FALSE;

              if ( tmppair->first->size() == 0 && tmppair->second->size() == 0 ) are_duplicates = TRUE;

              tmpfile = tmppair->first->brother;
              while( tmpfile && !are_duplicates )
              {
                if ( tmpfile == tmppair->second->brother ) are_duplicates = TRUE;
                tmpfile = tmpfile->duplicates;
              }

              tmpfile = tmppair->second->brother;
              while( tmpfile && !are_duplicates )
              {
                if ( tmpfile == tmppair->first->brother ) are_duplicates = TRUE;
                tmpfile = tmpfile->duplicates;
              }

              if ( !are_duplicates )  // if !size we shouldn't report all the files as different.
              {                       // the rare case of equal but same time isn't considered as user disabled search by size.
                if ( size )
                {
                  if ( enable_newer_search )
                  {
                    QPair<KFileItemExt*, KFileItemExt*> *resultpair = new QPair<KFileItemExt*, KFileItemExt*>;
                    *resultpair = qMakePair( tmppair->first, tmppair->second );

                    QCustomEvent *event = new QCustomEvent( RESULT_COMP_DIFFERENT );
                    event->setData( resultpair );
                    QApplication::postEvent( parent, event );
                  }
                }
                else // we don't search for size duplicates, and the file doesn't belong to missing or newer list.
                {    // so we abuse the duplicates list view to display the duplicates by path. (no duplicates by size will be displayed!).
                  if ( enable_duplicates_search )
                  {
                    tmppair->first->duplicates = tmppair->second;
                    tmppair->second->dup_parent = tmppair->first;
                    tmppair->first->hasdupes = 1;
                    tmppair->first->isdupe = 1;
                    tmppair->second->isdupe = 1;
                    QCustomEvent *event = new QCustomEvent( RESULT_DUPLICATE );
                    event->setData( tmppair->first );
                    QApplication::postEvent( parent, event );
                  }
                }
              }
            }
          }
        }

        QPair<KFileItemExt*, KFileItemExt*> *delpair;

        for ( delpair = list.first(); delpair; delpair = list.next() )
        {
          delete delpair;
        }
      }
      else
      {
        if ( !files_comp->isdupe_comp )
        {
          if ( enable_missing_search )
          {
            QCustomEvent *event = new QCustomEvent( RESULT_COMP_MISSING );
            event->setData( files_comp );
            QApplication::postEvent( parent, event );
          }
        }
      }

      current_file++;
      files_comp = files_comp->next;
    }
  }

  purgeTree( checktree_dupes );
  purgeTree( checktree_comp );

  if ( !cancel() )
  {
    QCustomEvent *event = new QCustomEvent( RESULT_FINISHED );
    QApplication::postEvent( parent, event );
  }

  initialized = FALSE;
}

int KompareJob::registerFile(fileTree **branch, KFileItemExt *file)
{
  *branch = new fileTree();

  (*branch)->file = file;
  (*branch)->left = NULL;
  (*branch)->right = NULL;

  return 1;
}

KFileItemExt *KompareJob::checkMatchDupes( fileTree **root, fileTree *checktree, KFileItemExt *file )
{
  if ( cancel() ) return NULL;

  int cmpresult;
  unsigned long fsize;

  // If inodes are equal one of the files is a hard link, which
  // is usually not accidental. We don't want to flag them as 
  // duplicates, unless the user specifies otherwise.

//   if (!ISFLAG(flags, F_CONSIDERHARDLINKS) && getinode(file->d_name) == 
//    checktree->file->inode) return NULL;
// FIXME: Hard link handling

  fsize = file->size();

  if ( fsize < checktree->file->size() ) cmpresult = -1;
  else if ( fsize > checktree->file->size() ) cmpresult = 1;
  else if ( calculate_checksum )
  {
    QCString file_md5 = file->MD5( (KomparatorWidget*)parent );
    if ( cancel() ) return NULL;
    QCString checktree_file_md5 = checktree->file->MD5( (KomparatorWidget*)parent );
    cmpresult = qstrcmp( file_md5, checktree_file_md5 );
  }
  else cmpresult = 0;

  if ( cmpresult < 0 )
  {
    if ( checktree->left != NULL )
    {
      return checkMatchDupes( root, checktree->left, file );
    }
    else
    {
      registerFile( &(checktree->left), file );
      return NULL;
    }
  }
  else if ( cmpresult > 0 )
  {
    if ( checktree->right != NULL )
    {
      return checkMatchDupes( root, checktree->right, file );
    }
    else
    {
      registerFile( &(checktree->right), file );
      return NULL;
    }
  }
  else return checktree->file;
}

KFileItemExt *KompareJob::checkMatchComp( fileTree **root, fileTree *checktree, KFileItemExt *file )
{
  if ( cancel() ) return NULL;

  int cmpresult;

  QString file_str, checktree_file_str;

  if ( file->dir == 0 )
    file_str = "." + file->url().url().right( file->url().url().length() - url1.url().length() );
  else
    file_str = "." + file->url().url().right( file->url().url().length() - url2.url().length() );

  if ( checktree->file->dir == 0 )
    checktree_file_str = "." + checktree->file->url().url().right( checktree->file->url().url().length() -
                                    url1.url().length() );
  else
    checktree_file_str = "." + checktree->file->url().url().right( checktree->file->url().url().length() -
                                    url2.url().length() );

  if ( file_str.length() < checktree_file_str.length() ) cmpresult = -1;
  else if ( file_str.length() > checktree_file_str.length() ) cmpresult = 1;
  else cmpresult = qstrcmp( file_str, checktree_file_str );

  if ( cmpresult < 0 )
  {
    if ( checktree->left != NULL )
    {
      return checkMatchComp( root, checktree->left, file );
    }
    else
    {
      registerFile( &(checktree->left), file );
      return NULL;
    }
  }
  else if ( cmpresult > 0 )
  {
    if ( checktree->right != NULL )
    {
      return checkMatchComp( root, checktree->right, file );
    }
    else
    {
      registerFile( &(checktree->right), file );
      return NULL;
    }
  }
  else return checktree->file;
}

bool KompareJob::confirmMatchDupes( KFileItemExt *item1, KFileItemExt *item2 )
{
  if ( cancel() ) return FALSE;

  mutex.lock();

  if ( item1->size() > binary_comparision_threshold*1048576 ||
       item2->size() > binary_comparision_threshold*1048576 )
  {
    mutex.unlock();
    return TRUE; // Don't perform binary comparision for files bigger than *** MB = *** *1048576 bytes.
  }

  mutex.unlock();

  bool ret = TRUE;

  QFile file1;
  file1.setName( item1->getFile( (KomparatorWidget*)parent ) );

  if ( cancel() )
  {
    item1->deleteFile( (KomparatorWidget*)parent );
    return FALSE;
  }

  QFile file2;
  file2.setName( item2->getFile( (KomparatorWidget*)parent ) );

  if ( file1.name().isEmpty() || file2.name().isEmpty() )
  {
    item1->deleteFile( (KomparatorWidget*)parent );
    item2->deleteFile( (KomparatorWidget*)parent );
    return TRUE;  // better return TRUE on failure to not lose possible duplicates.
  }

  if ( !file1.open( IO_Raw | IO_ReadOnly ) )
  {
    KMessageBox::error( 0, i18n( "Error opening file %1." ).arg( file1.name() ),
                        i18n( "Error while comparing files" ) );

    item1->deleteFile( (KomparatorWidget*)parent );
    item2->deleteFile( (KomparatorWidget*)parent );
    return TRUE;
  }

  if ( !file2.open( IO_Raw | IO_ReadOnly ) )
  {
    KMessageBox::error( 0, i18n( "Error opening file %1." ).arg( file2.name() ),
                        i18n( "Error while comparing files" ) );
    item1->deleteFile( (KomparatorWidget*)parent );
    item2->deleteFile( (KomparatorWidget*)parent );
    return TRUE;
  }

  if ( cancel() )
  {
    file1.close();
    file2.close();

    item1->deleteFile( (KomparatorWidget*)parent );
    item1->deleteFile( (KomparatorWidget*)parent );
    return FALSE;
  }

  uint buffer_size = 65536;
  int read_size1, read_size2;

  char buffer1[buffer_size];
  char buffer2[buffer_size];

  while ( ( read_size1 = file1.readBlock( buffer1, buffer_size ) ) > 0 )
  {
    if ( ( read_size2 = file2.readBlock( buffer2, buffer_size ) ) <= 0 )
    {
      ret = FALSE;
      break;
    }

    if ( cancel() )
    {
      file1.close();
      file2.close();

      item1->deleteFile( (KomparatorWidget*)parent );
      item1->deleteFile( (KomparatorWidget*)parent );
      return FALSE;
    }

    if ( qstrncmp( buffer1, buffer2, read_size1 ) != 0 || read_size1 != read_size2 )
    {
      ret = FALSE;
      break;
    }
  }

  if ( ret && file2.readBlock( buffer2, buffer_size ) > 0 ) ret = FALSE;

  file1.close();
  file2.close();

  item1->deleteFile( (KomparatorWidget*)parent );
  item2->deleteFile( (KomparatorWidget*)parent );

  return ret;
}

bool KompareJob::confirmMatchComp( KFileItemExt *item1, KFileItemExt *item2 )
{
  if ( cancel() ) return FALSE;

  QString item1_str, item2_str;

  if ( item1->dir == 0 )
    item1_str = "." + item1->url().url().right( item1->url().url().length() - url1.url().length() );
  else
    item1_str = "." + item1->url().url().right( item1->url().url().length() - url2.url().length() );

  if ( item2->dir == 0 )
    item2_str = "." + item2->url().url().right( item2->url().url().length() - url1.url().length() );
  else
    item2_str = "." + item2->url().url().right( item2->url().url().length() - url2.url().length() );

  if ( qstrcmp( item1_str, item2_str ) == 0 ) return TRUE;

  return FALSE;
}

KFileItemExt *KompareJob::newerItem(KFileItemExt *item1, KFileItemExt *item2)
{
  if ( cancel() ) return NULL;

  if ( item1->time( KIO::UDS_MODIFICATION_TIME ) > item2->time( KIO::UDS_MODIFICATION_TIME ) )
  {
    return item1; // bigger time means newer
  }
  else if ( item1->time( KIO::UDS_MODIFICATION_TIME ) < item2->time( KIO::UDS_MODIFICATION_TIME ) )
  {
    return item2;
  }
  else
  {
    return NULL; // both have equal timestamps.
  }
}

void KompareJob::purgeTree( fileTree *tree )
{
  if ( tree == NULL ) return;

  if ( tree->left != NULL ) purgeTree( tree->left );

  if ( tree->right != NULL ) purgeTree( tree->right );

  delete tree;
  tree = NULL;
}
